// requires JDK > 1.1.7
import java.util.Hashtable;

/** demonstrate conditional exception handling.
  */
public class Model {
  private static Hashtable model = new Hashtable ();	// name -> Model
  private String name;
  private IO io;

  public Model (String name, IO io) {
    this.name = name; this.io = io; model.put(name, this);
  }

  /** interactively manage invocations.
    */
  public void method (Activation caller) throws Activation {
    final class MyActivation extends Activation {
      final String stack;
      int sum;
      MyActivation (Activation caller) {
	super(caller);
        stack = caller != null ? ((MyActivation)caller).stack+name : name; 
      }
      /** interactively decide how the exception should be disposed of.
	*/
      public Object raise (Object info) throws Activation {
	loop: for (;;)
	  try {
            String word = io.inRaise(stack, name, sum);
	    try {
	      io.setSum(name, sum += Integer.parseInt(word));
	    } catch (NumberFormatException e) {
	      switch (word.charAt(0)) {
	      case 's':	info = super.raise(info); break;
	      case 'h':	handle(info);	// cannot return
	      default:	break loop;
              }
	    }
	  } catch (NullPointerException e) { }
	io.leaveRaise(name);
	return info;
      }
      public String toString () { return name; }
    };
    final MyActivation self = new MyActivation(caller);

    for (;;)
      try {
	String word = io.inMethod(self.stack, name, self.sum);
	try {
	  io.setSum(name, self.sum += Integer.parseInt(word));
	} catch (NumberFormatException e) {
	  Model m = (Model)model.get(word);
	  if (m != null)
	    m.method(self);
	  else if (word.charAt(0) == 'r')
	    self.raise(self.stack);
	  else
	    break;
	}
      } catch (Activation x) {
	if (x != self) throw x;
	io.handleInfo(name, self.sum, self.info());
      }
    io.leaveMethod(name);
  }
}
