#import "Model.h"

@interface MyActivation: Activation {
  Model *const model;
@public
  NSString *const stack;
  int sum;
}
- initWithModel:(Model*)model caller:(MyActivation*)caller;
@end

static NSMutableDictionary * model;			// name -> Model

@implementation Model
+ (void)initialize {
    model = [[NSMutableDictionary dictionary] retain];	// forever...
}

- initWithName:(NSString*)_name io:(id<IO>)_io {
    *(NSString**)&name = [_name copy];
    *(id*)&io = [_io retain];
    [model setObject:self forKey:name];			// forever retained...
    return self;
}

- (void)dealloc {
    [name release]; [io release];
    [super dealloc];
}

- (void)method:(Activation*)caller {
    MyActivation *const this = [[MyActivation alloc]
			initWithModel:self caller:(MyActivation*)caller];
    BOOL done = NO;
    do
      NS_DURING
	NSString * word = [io inMethod:this->stack :name :this->sum];
        Model * m = (Model*)[model objectForKey:word];
	int n;

        if (m != nil)
	  [m method:this];
	else if ([word characterAtIndex:0] == 'r')
	  [this raise:this->stack];
        else if ((n = [word intValue]) != 0)
	  [io setSum:name :this->sum += n];
	else
	  done = YES;
      NS_HANDLER
	if (localException != this) [localException raise];
	[io handleInfo:name :this->sum :[[this userInfo] objectForKey:@"info"]];
      NS_ENDHANDLER
    while (!done);
    [io leaveMethod:name];
    [this release];
}
@end

@interface Model (local)
- (NSString*)name;
- (id<IO>)io;
@end

@implementation Model (local)
- (NSString*)name {
    return name;
}

- (id<IO>)io {
    return io;
}
@end

@implementation MyActivation
- initWithModel:(Model*)_model caller:(MyActivation*)_caller {
    [super initWithCaller:_caller];
    *(Model**)&model = _model;
    *(NSString**)&stack = !_caller ? [model name] :
	[NSString stringWithFormat:@"%@%@", _caller->stack, [model name]];
    [stack retain];
    return self;
}

- (void)dealloc {
    [stack release];
    [super dealloc];
}

- raise:info {
    for (;;) {
      NSString * word = [[model io] inRaise:stack :[model name] :sum];
	int n;

      switch ([word characterAtIndex:0]) {
      case 's':	info = [super raise:info]; break;
      case 'h':	[self handle:info];	// cannot return
      default:	if ((n = [word intValue]) != 0)
		  [[model io] setSum:[model name] :sum += n];
		else {
		  [[model io] leaveRaise:[model name]];
		  return info;
		}
      }
    }
}
@end
